/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#include "OssAudioManager.hpp"
using namespace openstreetsound::tools;
using namespace openstreetsound::tools::audio;

OssAudioManager::OssAudioManager() {
	fprintf(stdout, "[OssAudioManager::OssAudioManager] Constructor\n");
	currentChannels = 0;
	playing = false;
}

OssAudioManager::~OssAudioManager() {
	fprintf(stdout, "[OssAudioManager::~OssAudioManager] Deconstructor\n");
	if (playing)
		stop();

	alutExit();
}

void OssAudioManager::init() {
	
	// Init openAL
	alutInit(0, NULL);
	// Clear Error Code (so we can catch any new errors)
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		#if _DEBUG
			printf("alutInit error: %d", error);
		#endif
	} else {
		#if _DEBUG
			fprintf(stdout, "[OssAudioManager::init] init ok!\n");
		#endif
	}
	alDistanceModel(AL_INVERSE_DISTANCE);
	//std::string str = ".\\res\\Sounds\\tearoots.wav";
	//ALbyte* path = (const_cast <char *> (str.c_str()));

	//loadSound(path, 1);

}

int OssAudioManager::loadSoundVector(std::vector<data::OssXmlResult*> results, double pinLon, double pinLat, float range) {
	
	// checking if there is a wave file for every channel
	std::vector<data::OssXmlResult*>::iterator it;
	std::string path;
	for (it = results.begin(); it != results.end(); ) {
		path = ".\\res\\Sounds\\" + (*it)->getType() + ".wav";
		if (checkExistency(path)) {
			it++;
		} else {
			#if _DEBUG
				printf("[OssAudioManager::loadSoundVector] unable to load %s\n", path.c_str());
			#endif
			it = results.erase(it);
		}
	}

	currentChannels = results.size();
	if (currentChannels > MAX_NUM_BUFFERS) {
		currentChannels = MAX_NUM_BUFFERS;
	}
	
	#if _DEBUG
		printf("[OssAudioManager::loadSoundVector] loading %i valid channels\n", currentChannels);
	#endif

	if (!prepareBuffers())
		return 1;

	// try to fill buffers
	for (int b = 0; b < currentChannels; b++) {
		// make path
		std::string str = ".\\res\\Sounds\\" + results.at(b)->getType() + ".wav";
		ALbyte* path = (const_cast <char *> (str.c_str()));
		// try to fill one buffer after another
		if (!fillBuffers(path, b)) {
			// filling the buffer failed
			alDeleteBuffers(currentChannels, buffers);
			return 2;
		}
	}

	if (!prepareSources()) {
		stop();
		return 3;
	}

	// fill the sources
	for (int s = 0; s < currentChannels; s++) {
		// try to fill one source after another
		if(!fillSources(s)) {
			// filling the sources failed
			alDeleteBuffers(currentChannels, buffers);
			alDeleteSources(currentChannels, sources);
			return 4;
		}
		/* setting the source position(s) x, y, and z (which is zero)*/
		alSource3f(sources[s], AL_POSITION, results.at(s)->getLon(), results.at(s)->getLat(), 0);
		/* the distance until the sound starts getting lower in volume.
			this particular distance and any closer will result in the same full volume*/		
		alSourcef(sources[s], AL_REFERENCE_DISTANCE, 0.00001);
		/* the farthest a sound can be away and will therefore be the most quiet
			I know that I am using an ellipse but the longest distance still is 0.0009 */
		alSourcef(sources[s], AL_MAX_DISTANCE, range);

		//float euclid = sqrt(pow(results.at(s)->getLon() - pinLon, 2) + pow(results.at(s)->getLat() - pinLat, 2));
		//printf("[OssAudioManager::loadSoundVector] euclid distance of source %i: %f\n", s, euclid);
	}

	/* setting the listener position */
	ALfloat pos[3] = {pinLon, pinLat, 0};
	alListenerfv(AL_POSITION, pos);
	
	/* 'at' vector (north) and 'up' vector (out of the screen ;-) ) */
	ALfloat ori[6] = {pinLon, pinLat + 0.0009, 0, 0, 0, 1};
	alListenerfv(AL_ORIENTATION, ori);
	
	#if _DEBUG
		printf("[OssAudioManager::loadSoundVector] finished loading %i channels\n", currentChannels);
	#endif
	return 0;
}

void OssAudioManager::start() {
	for (int s = 0; s < currentChannels; s++)
		alSourcePlay(sources[s]);

	playing = true;
}

void OssAudioManager::stop() {
	for (int s = 0; s < currentChannels; s++)
		alSourceStop(sources[s]);

	playing = false;
	alDeleteSources(currentChannels, sources);
	alDeleteBuffers(currentChannels, buffers);
}

void OssAudioManager::pause() {
	for (int s = 0; s < currentChannels; s++)
		alSourcePause(sources[s]);
}

bool OssAudioManager::prepareBuffers() {
	// Create the buffers
	alGenBuffers(currentChannels, buffers);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		#if _DEBUG
			printf("[OssAudioManager::prepareBuffers] alGenBuffers : %d\n", error);
		#endif
		return false;
	}
	return true;
}

bool OssAudioManager::prepareSources() {
	// Generate the sources
	alGenSources(currentChannels, sources);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		#if _DEBUG
			printf("[OssAudioManager::prepareSources] alGenSources : %d\n", error);
		#endif
		return false;
	}
	return true;
}

bool OssAudioManager::fillBuffers(ALbyte* soundpath, int position) {
	// load file into memory
	buffers[position] = alutCreateBufferFromFile(soundpath);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		#if _DEBUG
			printf("[OssAudioManager::fillBuffers] alutCreateBufferFromFile failed : %d\n", error);
		#endif
		return false;
	}
	#if _DEBUG
		printf("[OssAudioManager::fillBuffers] loaded WAVFile\n");
	#endif
	return true;
}

bool OssAudioManager::fillSources(int position) {
	alSourcei(sources[position], AL_BUFFER, buffers[position]);
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		#if _DEBUG
			printf("alSourcei : %d\n", error);
		#endif
		return false;
	}
	return true;
}

inline bool OssAudioManager::checkExistency(const std::string& name) {
	#if _DEBUG
		printf("[OssAudioManager::checkExistency] checking: %s\n", name.c_str());
	#endif
	struct stat buffer;   
	return (stat (name.c_str(), &buffer) == 0);
}

bool OssAudioManager::isPlaying() {
	return playing;
}