/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef OSSAUDIOMANAGER_HPP_
#define OSSAUDIOMANAGER_HPP_

#include <windows.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
#include <sys/stat.h>
#include <math.h>
#include "al.h"
#include "alc.h"
#include "AL/alut.h"

#include "../data/OssXmlResult.hpp"

#define MAX_NUM_BUFFERS 8
#define MAX_NUM_SOURCES 8

namespace openstreetsound {
	namespace tools {
		namespace audio {
			class OssAudioManager {
			public:
				
				/*
				OssAudioManager()

				Constructor
				sets the playing flag to false and the currentChannels to zero
				*/
				OssAudioManager();
				
				/*
				~OssAudioManager()

				Deconstructor
				stops the sounds and clears the sources
				calls the OpenAL deconstructor.
				*/
				~OssAudioManager();
				
				/*
				init()

				calls the ALUT init method to initialize OpenAL
				and sets the distance model
				*/
				void init();
				
				/*
				start()

				starts playing all the currently buffered sounds
				*/
				void start();
				
				/*
				stop()

				stops playing all the currently buffered sounds
				and frees the resources
				*/
				void stop();
				
				/*
				pause()

				pauses the playing of sounds
				*/
				void pause();
				
				/*
				loadSoundVector(std::vector<data::OssXmlResult*> results, double pinLon, double pinLat)

				Input: a vector made up of OssXmlResult elements. must not be loger then 8 items.
					and the location of the pin in logitude and latitude
				First every result will be checked if a local wave file is available.
				If not, the result will be dismissed.
				For each wave file a buffer will be filled and given to a source object.
				Every source object will recieve a position given from the longitude and latitude of the result object
				and then two volume attenuation distances will be set.
				Lastly the position and orientation of the listener object will be set.
				Returns false if any error occured.
				*/
				int loadSoundVector(std::vector<data::OssXmlResult*> results, double pinLon, double pinLat, float range);
				
				/*
				isPlaying()

				returns the playing flag.
				*/
				bool isPlaying();
				
			private:
				
				/*
				ALenum error

				An enumeration of error states of OpenAL
				*/
				ALenum error;
				
				/*
				ALuint buffers[]

				An array of OpenAl sound buffers.
				*/
				ALuint buffers[MAX_NUM_BUFFERS];
				
				/*
				ALuint sources[]

				An array of OpenAl sounce source objects.
				*/
				ALuint sources[MAX_NUM_SOURCES];
				
				/*
				int currentChannels

				The amount of currently loaded valid channels
				*/
				int currentChannels;
				
				/*
				bool playing

				a simple flag that is set true, whenever sounds is beeing played
				*/
				bool playing;
				
				/*
				bool prepareBuffers()

				A helpler funciton that justs prepares an amount of buffers to be filled
				dependet on the value of currentChannels
				Returns false if an error occured.
				*/
				bool prepareBuffers();
				
				/*
				bool prepareSources()

				A helpler funciton that justs prepares an amount of sources to be filled
				dependet on the value of currentChannels
				Returns false if an error occured.
				*/
				bool prepareSources();
				
				/*
				fillBuffers(ALbyte* path, int position)

				Fills the buffers. Is called in a loop in the loadSoundVector method.
				Will load a file, given by its previously checked path to an position in the buffers array.
				Returns false if an error occured.
				*/
				bool fillBuffers(ALbyte* path, int position);
				
				/*
				bool fillSources(int position)

				Fills the sources. Is called in a loop in the loadSoundVector mehtod.
				Will put a buffer by its given position into the source array on the same position.
				Returns false if an error occured.
				*/
				bool fillSources(int position);
				
				/*
				checkExistency(const std::string& name)

				Checks if a file by given path exits in the application files.
				*/
				inline bool checkExistency(const std::string& name);

			};
		}
	}
}
#endif