/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef OSSXMLREADER_HPP_
#define OSSXMLREADER_HPP_

#include <stdio.h>
#include <stdlib.h>     /* strtoll */
#include <math.h>
#include <string>
#include <libxml/xmlreader.h>
#include <map>
#include <vector>

#include "OssXmlResult.hpp"
#include "OssXmlComparator.hpp"


namespace openstreetsound {
	namespace tools {
		namespace data {
			/*
			Class OssXmlReader
			 
			XML Parsing class
			reads through the entire filtered.osm xml document
			and looks out for nodes that are close to the pin
			gets called by the OpenStreetSounds class
			*/
			class OssXmlReader {
			public:
				/*
				Constructor
				
				Fills the class variables lat, lon, id, key, val, ref,
				aSquared, bSquared, aDoubled and bDoubled with their content
				*/
				OssXmlReader();

				/*
				Destructor
				
				Calls the libxml clean up
				and empties the map oxresult and the vector oxrInWay
				*/
				~OssXmlReader();

				/*
				streamFile(const char *filename, float lon, float lat)
				
				Input: char * filename: the name of the file to be parsed
				(which must be a vaild and filtered osm xml file)
				float lon, lat: the location of the placed pin
				parses through the file and calls processNode() on each found node
				*/
				std::vector<OssXmlResult*> streamFile(const char *filename, float lon, float lat, float lonRange, float latRange);

				/*
				streamTest(const char *filename)
				
				Input: char * filename: the name of the file to be parsed
				parses through the file and calls testrunNode() on each found node
				use only for testung purposes. doesn't actually do anything with the found data
				*/
				void streamTest(const char *filename);

				/*
				checkFile(const char *filename)
				
				Input: char * filename: the name of the file to be checked for existence
				Return: true if file is found or
				false if file is not found
				gets called in the initialization phase of the OpenStreetSound class
				*/
				bool checkFile(const char *filename);

			private:
				/*
				float lon, lat
				
				the longitude and latitude of the OssPin
				*/
				float lat, lon;

				/*
				xmlChar *key, *val, *id, *ref
				
				pointers to:
				the key attribute of a "tag" element
				the value attribute of a "tag" element
				the id attribute of a "node" element
				the ref attribute of a "nd" element
				*/
				xmlChar *key, *val, *id, *ref;

				/*
				xmlTextReaderPtr reader
				
				an object from the libxml library which handles the parsing
				is basically a whole node at a time
				*/
				xmlTextReaderPtr reader;

				/*
				float aSquared, bSquared
				
				the ellipse coefficients to the power to slightly increased calculation speed
				*/
				float aSquared, bSquared;

				/*
				float aDoubled, bDoubled
				
				the ellipse coefficients doubled to slightly increased calculation speed
				*/
				float aDoubled, bDoubled;

				/*
				OssXmlResult found
				
				a temporary OssXmlResult element
				*/
				OssXmlResult *found;

				/*
				std::map<unsigned long, OssXmlResult*> oxresult
				
				a map with all the found OssXmlResult
				key (first): the node ids
				values (second): a pointer to the corresponding OssXmlResult object
				*/
				std::map<unsigned long, OssXmlResult*> oxresult;

				/*
				std::map<unsigned long, OssXmlResult*>::Iterator oxrIterator
				
				an Iterator for the map with all the found OssXmlResult
				*/
				std::map<unsigned long, OssXmlResult*>::iterator oxrIterator;

				std::multimap<float, OssXmlResult*> oxrExportable;
				std::multimap<float, OssXmlResult*>::iterator oxrExportableIterator;

				/*
				std::vector<OssXmlResult*> oxrInWay
				
				a vector with multiple OssXmlResults that have been found
				below a "way" element that share the same tag value
				*/
				std::vector<OssXmlResult*> oxrInWay;

				/*
				std::vector<OssXmlResult*>::iterator oxrI
				
				an Iterator for the vector with the OssXmlResults that share the same tag value
				*/
				std::vector<OssXmlResult*>::iterator oxrI;

				/*
				processNode(xmlTextReaderPtr reader)
				
				Input: a XmlTextReaderPtr which stores the current element in the xml file
				fills the oxresults map which will be given to the synthesizer
				if a node is found that is close to the OssPin, 
					a temporary OssXmlResult "found" is created 
					and it is stored in the oxresult map
				if a nd is found that shares the referenced id with an element in the oxresult map
					it is stored in the oxrInWay vector
				if a way is found both the found and the oxrInWys vector are deleted
				if a tag is found the value is read
					and stored in a found OssXmlResult
					or in all the elements of the oxrInWay vector
				*/				
				void processNode(xmlTextReaderPtr reader);

				/*
				testrunNode(xmlTextReaderPtr reader)
				
				a test method for parsing through the xml file
				use only for testung purposes. it wont save anything
				*/
				void testrunNode(xmlTextReaderPtr reader);
			};
		}
	}
}
#endif