/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#include "OssMap.hpp"

using namespace openstreetsound::tools::visual;

bool zoom_test = false;
//Tile* zoom_tile = new Tile("res/raster.png");
//Tile* first_tile = new Tile("res/firstTile.png");
//Tile* last_tile = new Tile("res/lastTile.png");
//Tile* origin_marker = new Tile("res/origin.png");

// costructor
// origin = left upper corner
OssMap::OssMap() {
	origin = new OssPoint();
	size = new OssPoint(512, 512);
	mapOffset = new OssPoint(1,0);
	
	zoom_factor = 0;
	tileF = new OssTileFetcher();
	firstTileIndex = new OssPoint();
	lastTileIndex = new OssPoint();
	firstTilePos = new OssPoint();
	lastTilePos = new OssPoint();

	tiles = new std::vector<std::vector<OssTile*>>();

	dragging = false;
}

OssMap::~OssMap() {
	delete origin;
	delete size;
	delete mapOffset;

	// index points and their position
	delete firstTileIndex;
	delete lastTileIndex;
	delete firstTilePos;
	delete lastTilePos;
	
	// Helper and 2d vector
	delete tileF;
	tiles->clear();
	delete tiles;
}

// draws the map onto the given window hdc
// reads the image (as of now out of a png file)
// draws it into the window
bool OssMap::draw(HDC hdc) {
	// TODO think about changing the iteration to this one:
	//std::vector< std::vector<OssTile*> >::const_iterator row;
	//std::vector<OssTile*>::const_iterator col; 
	//for (row = tiles->begin(); row != tiles->end(); ++row) {
	//	for (col = row->begin(); col != row->end(); ++col) {
	//		col->draw(hdc);
	//	}
	//}

	bool successful = true;

	int sizeX = tiles->size();
	int sizeY = tiles->at(0).size();

	Gdiplus::Graphics graphics(hdc);

	OssPoint boundaryX = tileF->getXBoundaries(zoom_factor);
	OssPoint boundaryY = tileF->getYBoundaries(zoom_factor);

	for (int i = 0; i < sizeX; i++) {
		for (int j = 0; j < sizeY; j++) {
			OssPoint pos = *origin + OssPoint(i*256-256, j*256-256);
			
			OssTile *ti = tiles->at(i).at(j);

			if (!ti->isValid() && !ti->isSubstituted()){
				SetCursor(LoadCursor(NULL, IDC_WAIT));
				if (ti->getTileX() >= boundaryX.getX() &&
					ti->getTileX() <= boundaryX.getY() &&
					ti->getTileY() >= boundaryY.getX() &&
					ti->getTileY() <= boundaryY.getY()) {
					#if _DEBUG
						fprintf(stdout, "[Map::generateTiles] downloading tile %i/%i/%i\n", ti->getTileZ(), ti->getTileX(), ti->getTileY());
					#endif
						if (!downloadTile(ti->getTileX(), ti->getTileY(), ti->getTileZ())) {
							successful = false;
						}
					ti->setValid();
				}
				SetCursor(LoadCursor(NULL, IDC_ARROW));
			}
			
			tiles->at(i).at(j)->draw(&graphics, pos);
			if (i == firstTileIndex->getX() && j == firstTileIndex->getY()) {
				// save the current coordinates into the first tile
				*firstTilePos = pos;
			}
			if (i == lastTileIndex->getX() && j == lastTileIndex->getY()) {
				// save the current coordinates into the last tile
				*lastTilePos = pos;
			}
		}
	}
	return successful;
}

bool OssMap::downloadTile(int x, int y, int z) {
	return tileF->cUrlFetch(z, x, y);
}

// returns the left upper corner of the image
OssPoint OssMap::getOrigin() {
	return *origin;
}

// returns the size of the image
OssPoint OssMap::getSize() {
	return *size;
}

// sets the origin to given point
void OssMap::setOrigin(OssPoint *p) {
	origin = p;
}

// moves the origin by pixels given from the point
void OssMap::moveOrigin(OssPoint p) {
	*origin += p;
	dragging = true;
}

// returns the index of the first tile
OssPoint OssMap::getFirstTile() {
	return *firstTileIndex;
}

// returns the index of the last tile
OssPoint OssMap::getLastTile() {
	return *lastTileIndex;
}

// moves the origin by given ratio in x and y
// TODO find out if still needed
void OssMap::moveOrigin(float ratio_x, float ratio_y, float multip) {
	int x = size->getX() * ratio_x * multip;
	int y = size->getY() * ratio_y * multip;

	#if _DEBUG
		fprintf(stdout, "[Map::moveorigin] size: %i %i\n", size->getX(), size->getY());
		fprintf(stdout, "[Map::moveorigin] movepoint ratio %f, %f\n", ratio_x, ratio_y);
		fprintf(stdout, "[Map::moveorigin] movepoint pixel %i, %i\n", x, y);
	#endif

	moveOrigin(OssPoint(x, y));
	dragging = true;
}

// zooms into or out of the map by given wheeldelta at wheelpos
void OssMap::zoom(int wheel_delta, OssPoint wheel_pos) {

	#if _DEBUG
		fprintf(stdout, "[Map::zoom] input wheel_delta: %i\n", wheel_delta);
		fprintf(stdout, "[Map::zoom] input wheel_pos: (%i, %i)\n", wheel_pos.getX(), wheel_pos.getY());
	#endif

	// detect if zooming is valid
	if (zoom_factor + wheel_delta < 0 || zoom_factor + wheel_delta > 5) {
		#if _DEBUG
			fprintf(stdout, "[Map::zoom] zoom limit reached (%i)\n", zoom_factor);
		#endif
		return;
	}

	// the location of the mousewheel in long/lat
	std::pair<double, double> location = calculatePoint(wheel_pos);

	zoom_factor += wheel_delta;

	generateTiles();
	centerOn(location.first, location.second, wheel_pos);
	generateTiles();

	#if _DEBUG
		fprintf(stdout, "[Map::zoom] new zoom level: %i\n", zoom_factor);
	#endif
}

// generates the Tile based 2d vector and fills it with the right tiles
void OssMap::generateTiles() {
	bool firstFound = false;

	/* When the user is dragging the map around, lets only repaint it, when we really need to
		because the tiles start to be "out of view" */
	if (dragging) {
		dragging = false;
		if (origin->inBetween(OssPoint(-256, -256), OssPoint(256, 256)))
			return;
		#if _DEBUG
			fprintf(stdout, "[Map::generateTiles] generating tiles: Origin: %i : %i\n", origin->getX(), origin->getY());
		#endif
	}
	/* comment the above block to enable constant window refreshing*/

	OssPoint tilesSpace = *size/256 + OssPoint(3,3); // + (2,2) so it draws over the window edges

	OssPoint boundaryX = tileF->getXBoundaries(zoom_factor);
	OssPoint boundaryY = tileF->getYBoundaries(zoom_factor);
	//OssPoint available = OssPoint(boundaryX.getY()-boundaryX.getX() +1, boundaryY.getY()-boundaryY.getX() +1);

	/* calculating an offset to know which tile to start drawing at */
	OssPoint change = OssPoint(origin->getX() / 256, origin->getY() / 256);
	*mapOffset += change;
	origin->setX(origin->getX() % 256);
	origin->setY(origin->getY() % 256);

	/* freeing up memory, by deleting the previous pointers! */
	/* vector.clear() sadly does not call the deconstructors of the elements */
	std::vector< std::vector<OssTile*> >::const_iterator row;
	std::vector<OssTile*>::const_iterator col; 
	for (row = tiles->begin(); row != tiles->end(); ++row) {
		for (col = row->begin(); col != row->end(); ++col) {
			delete *col;
		}
	}
	tiles->clear();
	tiles->resize(tilesSpace.getX(), std::vector<OssTile*>(tilesSpace.getY(), new OssTile()));

	/* filling the tiles 2d vec */
	for (int i = 0; i < tilesSpace.getX(); i++) {
		for (int j = 0; j < tilesSpace.getY(); j++) {
			int x,y,z;
			z = (zoom_factor +12);
			/* starting at boundry + offset */
			x = boundaryX.getX() + i - mapOffset->getX();
			y = boundaryY.getX() + j - mapOffset->getY();
			tiles->at(i).at(j) = new OssTile(z, x, y);
			
			/* finding the first and last tile with an actual map tile */
			if (tiles->at(i).at(j)->isValid()) {
				lastTileIndex = new OssPoint(i,j);
				if (!firstFound) {
					firstTileIndex = new OssPoint(i,j);
					firstFound = true;
				}
			}
		}
	}
}

// centers the given coordinates onto the point in window coordinates
void OssMap::centerOn(double longitude, double latitude, OssPoint c) {
	#if _DEBUG
		fprintf(stdout, "[Map::centerOn] long, lat: %f, %f\n", longitude, latitude);
		fprintf(stdout, "[Map::centerOn] center point: %i : %i\n", c.getX(), c.getY());
	#endif
	int z = zoom_factor+12;

	int tilelux = tileF->long2tilex(longitude, z);
	int tileluy = tileF->lat2tiley(latitude, z);

	double tileLong = tileF->tilex2long(tilelux, z);
	double tileLat = tileF->tiley2lat(tileluy, z);
	double nextTileLong = tileF->tilex2long(tilelux+1, z);
	double nextTileLat = tileF->tiley2lat(tileluy+1, z);

	double longLength = nextTileLong - tileLong; // length of a tile in long
	double latLength = nextTileLat - tileLat;	// length of a tile in lat

	std::pair<double, double> longlat = calculatePoint(c); // the long and lat of the focal point
	double distanceLong = longitude - longlat.first; // difference in long of the POI and the focal Point
	double distanceLat = latitude - longlat.second;
	
	int distanceX = (int) (distanceLong / longLength * 256);  // distance in pixel
	int distanceY = (int) (distanceLat / latLength * 256);

	moveOrigin(OssPoint(-distanceX, -distanceY));

	#if _DEBUG
		fprintf(stdout, "[Map::centerOn] distance x,y: %i, %i\n", distanceX, distanceY);
		fprintf(stdout, "[Map::centerOn] distance longlat: %f, %f\n", distanceLong, distanceLat);
	#endif
}

// gives the window size to the map
void OssMap::setWindowSize(OssPoint *p) {
	*size = *p;
	
	generateTiles();
}

// calculates the long and lat of a given point in window coordinates
std::pair<double, double> OssMap::calculatePoint(OssPoint p) {
	
	double distance_x = p.getX() - firstTilePos->getX();
	double distance_y = p.getY() - firstTilePos->getY();
	OssPoint pTile = OssPoint(distance_x/256, distance_y/256) + *firstTileIndex;

	double ratio_x = ((double) ((int) distance_x % 256)) / 256;
	double ratio_y = ((double) ((int) distance_y % 256)) / 256;

	OssTile& tile = *tiles->at(pTile.getX()).at(pTile.getY());
	int tx = tile.getTileX();
	int ty = tile.getTileY();
	int tz = tile.getTileZ();

	double tileLong = tileF->tilex2long(tx, tz);
	double tileLat = tileF->tiley2lat(ty, tz);
	double nextTileLong = tileF->tilex2long(tx+1, tz);
	double nextTileLat = tileF->tiley2lat(ty+1, tz);

	// tile lenght * ratio
	double offsetLong = (nextTileLong - tileLong) * ratio_x;
	double offsetLat = (nextTileLat - tileLat) * ratio_y;

	double pointLong = tileLong + offsetLong;
	double pointLat = tileLat + offsetLat;

	#if _DEBUG
		fprintf(stdout, "[Map::calculatePoint] First Tile Pos: %i, %i\n", firstTilePos->getX(), firstTilePos->getY());
		fprintf(stdout, "[Map::calculatePoint] First Tile X, Y: %i, %i\n", firstTileIndex->getX(), firstTileIndex->getY());
		fprintf(stdout, "[Map::calculatePoint] Distances: %i, %i\n", distance_x, distance_y);
		fprintf(stdout, "[Map::calculatePoint] Input Point: %i, %i\n", p.getX(), p.getY());
		fprintf(stdout, "[Map::calculatePoint] Tile Info: %i/%i/%i\n", tz, tx, ty);
		fprintf(stdout, "[Map::calculatePoint] Point Long and Lat: %f, %f\n", pointLong, pointLat);
	#endif

	return std::make_pair(pointLong, pointLat);
}

// calculates the pixel position on the current map of a point given in long and lat
OssPoint OssMap::calculateLongLat(double longitude, double latitude) {
	OssTile& lu = *tiles->at(firstTileIndex->getX()).at(firstTileIndex->getY());
	OssTile& rl = *tiles->at(lastTileIndex->getX()).at(lastTileIndex->getY());
	
	double lulong = tileF->tilex2long(lu.getTileX(), lu.getTileZ());
	double lulat = tileF->tiley2lat(lu.getTileY(), lu.getTileZ());

	double rllong = tileF->tilex2long(rl.getTileX(), rl.getTileZ()); // 
	double rllat = tileF->tiley2lat(rl.getTileY(), rl.getTileZ());
	
	double lengthlong = rllong - lulong;
	double lengthlat = rllat - lulat;

	double distlong = longitude - lulong;
	double distlat = latitude - lulat;

	double ratiolong = distlong / lengthlong;
	double ratiolat = distlat / lengthlat;

	int lengthx = rl.getPosition().getX() - lu.getPosition().getX();
	int lengthy = rl.getPosition().getY() - lu.getPosition().getY();

	int distx = (int) (lengthx * ratiolong) + lu.getPosition().getX();
	int disty = (int) (lengthy * ratiolat) + lu.getPosition().getY();

	#if _DEBUG
		fprintf(stdout, "[Map::calculateLongLat] Input LongLat: %f, %f\n", longitude, latitude);
		fprintf(stdout, "[Map::calculateLongLat] LU LongLat/TileXY: %f, %f / %i, %i\n", lulong, lulat, lu.getPosition().getX(), lu.getPosition().getY());
		fprintf(stdout, "[Map::calculateLongLat] RL LongLat/TileXY: %f, %f / %i, %i\n", rllong, rllat, rl.getPosition().getX(), rl.getPosition().getY());
		fprintf(stdout, "[Map::calculateLongLat] Ratios: %f, %f\n", ratiolong, ratiolat);
		fprintf(stdout, "[Map::calculateLongLat] Point x, y: %i, %i\n", distx, disty);
	#endif
	
	return OssPoint(distx, disty);
}

void OssMap::setServer(std::string _server) {
	tileF->setServer(_server);
}

void OssMap::findSubstituted() {
	std::vector< std::vector<OssTile*> >::const_iterator row;
	std::vector<OssTile*>::const_iterator col; 
	for (row = tiles->begin(); row != tiles->end(); ++row) {
		for (col = row->begin(); col != row->end(); ++col) {
			(*col)->invalidateSubstitude();
		}
	}
}

OssPoint OssMap::getTileFBoundriesX(int zoomfactor) {
	return tileF->getXBoundaries(zoomfactor);
}

OssPoint OssMap::getTileFBoundriesY(int zoomfactor) {
	return tileF->getYBoundaries(zoomfactor);
}

void OssMap::printTileBorders() {
	/*
	Border checking:
	*/

	double longDown = 16.5786;
	double longUp = 16.1827;
	double latRight = 48.1183;
	double latLeft = 48.3231;

	#if _DEBUG
		fprintf(stdout, "[Map::Map] Corners: \n");
		fprintf(stdout, "[Map::Map] Left Upper Tile for zoom 12: x=%i, y=%i\n", tileF->long2tilex(longUp, 12), tileF->lat2tiley(latLeft, 12));
		fprintf(stdout, "[Map::Map] Left Upper Tile for zoom 17: x=%i, y=%i\n", tileF->long2tilex(longUp, 17), tileF->lat2tiley(latLeft, 17));
	#endif
}