/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef OSSTILE_HPP_
#define OSSTILE_HPP_

/* STL & Windows headers */
#include <string>
#include <windows.h>
#include <iostream>
#include <fstream>
#include <gdiplus.h>
#include <gdiplusheaders.h>

/* own header */
#include "OssPoint.hpp"

namespace openstreetsound {
	namespace tools {
		namespace visual {
			/*
			Class OssTile
			
			a map tile which is used to store information about a tile
			stores the png picture and draws it on call
			*/
			class OssTile {
			public:
				/* 
				Constructor
				
				creates a simple but useless Tile which is not valid
				has the default png and invalid coordinates
				*/
				OssTile();

				/* 
				Copy constructor
				
				Input: another tile
				creates a deep copy of the given tile 
				*/
				OssTile(const OssTile& other);
				
				/* 
				Debug Constructor

				Input: std:string path for downloading the tile
				Only use this for debug uses
				create a tile with invalid coordinates but with a valid path
				so one can test downloading that tile (or whatever else the path is)
				the path will be added to the server path in the TileFetcher if dl is requested
				*/
				OssTile(std::string path);
				
				/* 
				Costructor (regular)
				
				Input: integer x, y and z
				tile coordinates in terms of "./res/Tiles/z/x/y.png"
				stores the coordinates and tries to find a local tile (.png file)
				with given coordinates. If found stores the file into the image pointer
				*/
				OssTile(int z, int x, int y);

				/*
				Deconstructor
				
				deletes the position pointer (OssPoint)
				and the image pointer (Gdiplus::Image)
				*/
				~OssTile();

				/*
				draw(Gdiplus::Graphics *graphics, OssPoint p)
				
				Input: Gdiplus Graphics, given by the OpenStreetSound class
				is a "pointer to the canvas on which the program draws"
				OssPoint p the window position where the image should be drawn
				gets called by the map, for every tile
				tells the tile to draw its image pointer at the given position
				*/
				void draw(Gdiplus::Graphics *graphics, OssPoint p);

				/*
				isValid()
				
				getter for the boolean flag vaild which tells if a png file
				for the coordinates given by to the constructor was found by it
				*/
				bool isValid();

				/* 
				setValid()
				
				tries to find a png file with the saved coordinates
				if one is found, sets valid to true,
				and fills the image pointer with the png
				sould only be used if a file which has previously been not vaild
				has been downloaded in the meantime, and now should be a valid tile.
				*/
				void setValid();

				/*
				invalidate()
				
				sets the tile to invalid if the tile is subtituted
				*/
				void invalidateSubstitude();

				/*
				isSubstituted()
				
				returns true, if the program is using parts of a higher zoom pic
				to substitute the tile.
				*/
				bool isSubstituted();

				/*
				getPosition()
				
				getter to the OssPoint pointer position
				which stores the window coordinates
				*/
				OssPoint getPosition();

				/*
				getPath()
				
				returns the std::string path,
				which is the local path
				looks like "./res/Tiles/z/x/y.png" for a valid tile
				or ".res/stone.png" for an invalid
				*/
				std::string getPath();

				/*
				getTileZ()
				 
				getter for the integer tile z coordinate
				*/
				int getTileZ();

				/*
				getTileX()
				 
				getter for the integer tile x coordinate
				*/
				int getTileX();

				/*
				getTileY()
				 
				getter for the integer tile y coordinate
				*/
				int getTileY();

			private:
				/*
				integers tileZ, tileX, tileY
				
				store the coordinates of the tile
				*/
				int tileZ, tileX, tileY;

				/* 
				string path
				
				stores the local path of the tile
				*/
				std::string path;

				/*
				OssPoint* position
				
				stores the position of the tile in window coordinates
				gets updated every call of this.draw()
				*/
				OssPoint *position;

				/*
				OssPoint* start
				
				stores the coordinates on the image where to start drawing
				is usually 0,0, only different for some images that couldn't be downloadad
				then its either (0,0) or (0,128) or (128,0) or (128,128)
				*/
				OssPoint *start;

				/*
				OssPoint* start
				
				stores the coordinates on the image where to end drawing
				is usually 256,256, only different for some images that couldn't be downloadad
				then its either (256,256) or (256,128) or (128,256) or (128,128)
				*/
				OssPoint *end;

				/*
				boolean valid
				
				is true in case a png file with the tiles coordinates has been found locally
				*/
				bool valid;

				bool substituted;

				/*
				fexists(const char *filename)
				
				checks if a file exits locally
				*/
				bool fexists(const char *filename);

				void findSubstitute(int factor, int prevOffsetx, int prevOffsetY);

				/*
				Gdiplus::Image (pointer) tileImg
				
				stores the png file for faster drawing
				*/
				Gdiplus::Image *tileImg;
			};
		}
	}
}

#endif
